@@ -2,7 +2,7 @@ import random
 import time
 from concurrent.futures import ThreadPoolExecutor
 from queue import Queue, Empty
-from threading import Lock
+from threading import Lock, Condition, RLock
 from typing import List, Callable, Optional, Any
 
 import requests
@@ -14,6 +14,41 @@ from simple_proxy2.tools.random_user_agent import get_random as random_agent
 from simple_proxy2.tools.simple_timer import SimpleTimer
 
 
+class Task:
+    def __init__(self, fn: Callable):
+        self._fn = fn
+
+        self._result = None
+        self._started = False
+        self._done = False
+        self._mutex = RLock()
+        self._condition = Condition(self._mutex)
+
+    def get(self):
+        with self._mutex:
+            if self._done:
+                return
+
+            self._condition.wait()
+            return self._result
+
+    def __call__(self, *args):
+        with self._mutex:
+            if self._started:
+                return
+
+            self._started = True
+
+            assert not self._done, "task is already finished"
+
+            try:
+                self._result = self._fn(*args)
+                self._done = True
+                self._condition.notify_all()
+            except Exception as ex:
+                print(ex)
+
+
 class ProxyManager:
     def __init__(self,
                  test_url: str,
@@ -61,12 +96,29 @@ class ProxyManager:
             else:
                 return self._success / self._trials
 
+    def proxy_as_request_session(self) -> Session:
+        return SessionProxy(self)
+
     def proxy_session(self,
                       fn: Callable[[Session, tuple], Optional[Any]],
                       callback_async: Callable[[Optional[Any]], None] = lambda x: None,
                       exception_handle: Callable[[ProxyInfo, Exception], None] = lambda info, ex: print(info, ex),
-                      args: tuple = ()):
-        def task():
+                      args: tuple = (),
+                      wait=False):
+        """
+        Work on the given function using the randomly chosen Session from the proxy pool. 'fn' will be retried
+        indefinitely until success in order to try all available proxies.
+
+        :param fn: function to execute. If has arguments, put those in the 'args'
+        :param callback_async: callback to be invoked after 'fn' is done. Return value from 'fn' will be used as
+         parameters
+        :param exception_handle: decide what to do when exception occurred
+        :param args: arguments to pass to 'fn'
+        :param wait: block the thread and wait for the 'fn' to finish.
+        :return: if 'wait' is True, the return value of 'fn' will be returned; None otherwise
+        """
+
+        def task_fn():
             success = False
 
             while not success:
@@ -87,6 +139,7 @@ class ProxyManager:
                             callback_async(result)
 
                         success = True
+                        return result
                     except Exception as ex:
                         exception_handle(proxy.info(), ex)
                     finally:
@@ -100,8 +153,12 @@ class ProxyManager:
                         if self._verbose:
                             print("success rate:", self.success_rate())
 
+        task = Task(task_fn)
         self._executor_queue.put(task)
 
+        if wait:
+            return task.get()
+
     def start(self):
         self._pool.start()
 
@@ -135,3 +192,27 @@ class ProxyManager:
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.stop()
+
+
+class SessionProxy(Session):
+    def __init__(self, manager: ProxyManager):
+        super().__init__()
+        assert manager._executor_running, "manager hasn't started."
+
+        self._manager = manager
+
+    def request(self, method, url,
+                params=None, data=None, headers=None, cookies=None, files=None,
+                auth=None, timeout=None, allow_redirects=True, proxies=None,
+                hooks=None, stream=True, verify=None, cert=None, json=None):
+
+        def fn(session: Session):
+            if headers is not None:
+                session.headers.update(headers)
+
+            return session.request(method, url,
+                                   params, data, headers, cookies, files,
+                                   auth, timeout, allow_redirects, None,
+                                   hooks, stream, verify, cert, json)
+
+        return self._manager.proxy_session(fn, wait=True)
